function htmlencode(string) 
	return gsub(string, ' ', '+')
end

function strjoinarray(sep, array)
	assert(#array > 0)
	if (#array == 1) then return array[1]
	else
		local result = array[1]
		for i=2,#array do
			result = result..sep..array[i]
		end
		return result;
	end
	assert(false)
end

CHARDB_BLIZZARD2SIMCRAFT_WEAPONSTATS = {
	["DPS_TEMPLATE"]="DPS",
--	["DAMAGE_TEMPLATE_WITH_SCHOOL"]="damage",
--	["DAMAGE_TEMPLATE"]="damage",
	["WEAPON_SPEED"]="Speed",
}

CHARDB_BLIZZARD2SIMCRAFT_STATS = {
	["ITEM_MOD_AGILITY"]="Agi", 
	["ITEM_MOD_SPIRIT"] = "Spi",
	["ITEM_MOD_STAMINA"] = "Sta", 
	["ITEM_MOD_STRENGTH"] = "Str", 
	["ITEM_MOD_INTELLECT"] = "Int", 

	["ITEM_MOD_SPELL_POWER"] = "SP",
	["ITEM_MOD_MANA_REGENERATION"] = "MP5",
	["ITEM_MOD_SPELL_PENETRATION"] = "SPen",

	["ITEM_MOD_ATTACK_POWER"] = "AP",
	["ITEM_MOD_EXPERTISE_RATING"] = "Exp",
	["ITEM_MOD_ARMOR_PENETRATION_RATING"] = "APen",
	
	["ARMOR_TEMPLATE"] = "Armor",
--	["ITEM_MOD_RESILIENCE_RATING"] = "Resil",

	["ITEM_MOD_PARRY_RATING"] = "Parry",
	["ITEM_MOD_DODGE_RATING"] = "Dodge",
	["ITEM_MOD_BLOCK_RATING"] = "Block",
	["ITEM_MOD_DEFENCE_RATING"] = "Def",

	["ITEM_MOD_CRIT_RATING"] = "Crit",
	["ITEM_MOD_HASTE_RATING"] = "Haste",
	["ITEM_MOD_HIT_RATING"] = "Hit",

}

CHARDB_ARMORY_CLASS_IDS = {
	WARRIOR = 1,
	PALADIN = 2,
	HUNTER = 3,
	ROGUE = 4,
	PRIEST = 5,
	DEATHKNIGHT = 6,
	SHAMAN = 7,
	MAGE = 8,
	DRUID = 11
}
CHARDB_SIMCRAFT_SERVERLOCATION = "region=us,server="..htmlencode(GetRealmName()) -- FIXME region

CHARDB_INVENTORY_IDS = {
	[01]='head',
	[02]='neck',
	[03]='shoulder',
	[04]='shirt',
	[05]='chest',
	[06]='waist',
	[07]='legs',
	[08]='feet',
	[09]='wrists',
	[10]='hands',
	[11]='finger1',
	[12]='finger2',
	[13]='trinket1',
	[14]='trinket2',
	[15]='back',
	[16]='main_hand',
	[17]='off_hand',
	[18]='ranged',
	[19]='tabard'
}

local itemData = LibItemDB.GetItemData;

CharDBUnit = {}
CharDBUnit.__index = CharDBUnit;

function CharDBUnit:New(unitId)
	local object = {};
	setmetatable(object, self);
	object.unitId = unitId;
	object.GUID = UnitGUID(unitId);
	object.name = UnitName(unitId);
	object.classLocal, object.class = UnitClass(unitId);
	object.level = UnitLevel(unitId);
	object:ResetGear();
	object:ResetTalents();
	return object;
end

function CharDBUnit:Complete()
	return self:GearComplete() and self:TalentsComplete();
end

function CharDBUnit:GearComplete()
	return self.gear ~= nil and self.gearPartial ~= true;
end

function CharDBUnit:TalentsComplete()
	return self.talentGroups ~= nil;
end

function CharDBUnit:Outdated(time)
	return self:GearOutdated(time) or self:TalentsOutdated(time);
end

function CharDBUnit:GearOutdated(time)
	return self.gearDate < time;
end

function CharDBUnit:TalentsOutdated(time)
	return self.talentsDate < time;
end

function CharDBUnit:SetGear(gear, gearPartial)
	self:ResetGear()
	self.gear = gear;
	self.gearPartial = gearPartial;
	self.gearDate = time();
end

function CharDBUnit:ResetGear()
	self.gearPartial = false;
	self.gear = nil;
	self.gearDate = 0;
	self.sets = nil;
	self.metaGem = nil;
end

function CharDBUnit:SetTalents(talentGroups, activeTalentGroup)
	self:ResetTalents();
	self.activeTalentGroup = activeTalentGroup;
	self.talentGroups = talentGroups;
	self.talentsDate = time();
end

function CharDBUnit:ResetTalents()
	self.talentGroups = nil;
	self.activeTalentGroup = nil;
	self.talentsDate = 0;
end

function CharDBUnit:ValidateUnitId() 
	local result = self.unitId ~= nil and UnitGUID(self.unitId) == self.GUID
	if (not result) then self.unitId = nil end
	return result;
end

function CharDBUnit:GetTalentURL() 
	if (not self:TalentsComplete()) then return nil end;
	local url = 'h'..'ttp://www.wowarmory.com/talent-calc.xml?cid='..CHARDB_ARMORY_CLASS_IDS[self.class].."&tal="
	local talents = self:GetTalents();
	for i=1,#talents do
		for j=1, #talents[i] do
			url = url..talents[i][j]
		end
	end
	return url;
end

function CharDBUnit:GetSpec(...) 
	if (not self:TalentsComplete()) then return nil end;
	local talentsGroup = self.activeTalentGroup
	if (select("#", ...) > 0) then
		talentsGroup = select(1, ...)
	end	local spec = {}
	local talents = self:GetTalents(talentsGroup)
	for i=1,3 do
		spec[i] = 0
		for j=1,#talents[i] do
			spec[i] = spec[i] + talents[i][j]
		end
	end
	return spec;
end

function CharDBUnit:GetMetaGem()
	if (not self:GearComplete()) then return nil end;
	if (self.metaGem == nil) then
		local helm = itemData(self.gear[1]);
		--assert(#helm.GEMS == 4)
		--for i=1,4 do
		for _, gemId in pairs(helm.GEMS) do
--			local gemId = helm.gemIds[i]
			for metaGemId, metaGemId in pairs(CHARDB_METAGEMS) do
				if (gemId == metaGemId) then
					self.metaGem = metaGemId;
				end
			end
		end
	end
	return self.metaGem
end

function CharDBUnit:GetSets()
	if (not self:GearComplete()) then return nil end;
	if (self.setBonuses == nil) then
		local setData = CHARDB_SETS[self.class]
		local sets = {}
		for setName, set in pairs(setData) do
			sets[setName] = 0;
			for _, item in pairs(self.gear) do
				for _, itemID in pairs(set) do
					if (itemID == itemData(item).itemId) then
						sets[setName] = sets[setName] + 1;
					end
				end
			end
		end
		self.sets = sets;
	end
	return self.sets
end

function CharDBUnit:GetTalents(...)
	if (not self:TalentsComplete()) then return nil end;
	local talentsGroup = self.activeTalentGroup
	if (select("#", ...) > 0) then
		talentsGroup = select(1, ...)
	end
	return self.talentGroups[talentsGroup];
end

function CharDBUnit:GetSpecName(...) 
	if (not self:TalentsComplete()) then return nil end;
	local talentsGroup = self.activeTalentGroup
	if (select("#", ...) > 0) then
		talentsGroup = select(1, ...)
	end
	local spec = {}	
	local max = 0;
	local maxIndex = 0;
	local spec = self:GetSpec(talentsGroup)
	if (spec == nil) then return nil end;
	for tabIndex, count in pairs(spec) do
		if (count > max) then
			maxIndex = tabIndex
			max = count
		end
	end
	if (max == 0) then
		return ''
	else
		return CHARDB_TALENT_TAB_NAMES[self.class][maxIndex];
	end
end

function CharDBUnit:GetSimcraftParams()

	-- Do basic data -- this is a armory player option!
	local text = "player="..self.name..","..CHARDB_SIMCRAFT_SERVERLOCATION..',talents='..self:GetSpecName()..'\n'
	text = text.."level="..self.level..'\n';
	text = text.."talents="..self:GetTalentURL() ..'\n';
	
	-- Do gear
	for i, link in pairs(self.gear) do
		text=text..CHARDB_INVENTORY_IDS[i]..'='..self.GetItemString(link)..'\n';
	end
	
	-- Do set bonuses
	local setBonuses = self:GetSets();
	for setName, count in pairs(setBonuses) do
		if (count >= 2) then
			text = text..setName.."_2pc=1"..'\n'
			if (count >= 4) then
				text = text..setName.."_4pc=1"..'\n'
			else
				text = text..setName.."_4pc=0"..'\n'
			end
		else
			text = text..setName.."_2pc=0"..'\n'
			text = text..setName.."_4pc=0"..'\n'
		end
	end
	
	-- Do meta gem
	local metaGem = self:GetMetaGem()
	if (metaGem ~= nil) then
		text = text.."meta_gem="..metaGem..'\n'
	end
	
	-- Do "unique" item (items with proc or on use effects)
	local uniqueItems = {}
	for _, itemLink in pairs(self.gear) do
		local name = SimcraftizeName(itemData(itemLink).name)
		for _, uniqueItemName in pairs(CHARDB_UNIQUEITEMS) do
			if (name == CHARDB_UNIQUEITEMS) then
				tinsert(uniqueItems, name)
			end
		end
	end
	if (#uniqueItems > 0) then
		local uniqueItemsTest = strjoinarray('/',  uniqueItems)
		text = text..'items='..uniqueItemsTest..'\n'
	end

	return text;
end

function CharDBUnit.GetItemString(link) 
	local item = itemData(link);
	local string = SimcraftizeName(item['INFO'][1]);
	--if (#item.stats > 0) then
	local statString = "";
	for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do
		if (item[blizName]) then 
			statString = statString..item[blizName]..simName..'_';
		end
	end
	if (strlen(statString) ~= 0) then
		statString = strsub(statString, 0, strlen(statString)-1);
		string = string..",stats="..statString;
	end
	
	if(item["INVTYPE_RANGED"] or item["INVTYPE_2HWEAPON"] or item["INVTYPE_WEAPON"] or item["INVTYPE_WEAPONMAINHAND"] or item["INVTYPE_WEAPONOFFHAND"]) then
		--assert(#item.weaponStats == 2)
		local subtype = LIBITEMDB_ITEMTYPES[item['INFO'][7]];
		if (subtype ~= 'Wand' and simtype ~= nil) then
			subtype = SimcraftizeName(subtype);
			subtype = gsub(subtype, 'twohanded_', '');
			subtype = gsub(subtype, 'onehanded_', '');
			string = string..",weapon="..subtype..'_';
		
			for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_WEAPONSTATS) do
				if (item[blizName]) then 
					string = string..item[blizName]..simName..'_';
				end
			end
			string = strsub(string, 0, strlen(string)-1);
		end
	end
	
	if (item["ENCHANT"]) then
		local enchantData = LibItemDB.GetEnchantData(item["ENCHANT"]);
		local enchantString = "";
		for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do
			if (enchantData[blizName]) then 
				enchantString = enchantString..enchantData[blizName]..simName..'_';
			end
		end
		if (strlen(enchantString) ~= 0) then
			string = string..",enchant="..strsub(enchantString, 0, strlen(enchantString)-1);
		end
	end

	
	if (item["GEMS"] and #item["GEMS"] ~= 0) then
		
		local gemsData = {};
		for i,gem in pairs(item["GEMS"]) do
			local gemData = LibItemDB.GetGemData(gem);
			for name, value in pairs(gemData) do
				if (type(value) == 'number') then
					if (not gemsData[name]) then gemsData[name] = 0 end
					gemsData[name] = gemsData[name] + value;
				end
			end
		end
		
		if (item["ITEM_SOCKET_BONUS"]) then
			for name, value in pairs(item["ITEM_SOCKET_BONUS"]) do
				if (type(value) == 'number') then
					if (not gemsData[name]) then gemsData[name] = 0 end
					gemsData[name] = gemsData[name] + value;
				end
			end
		end
		local gemsString = ""
		for blizName, simName in pairs(CHARDB_BLIZZARD2SIMCRAFT_STATS) do
			if (gemsData[blizName]) then 
				gemsString = gemsString..gemsData[blizName]..simName..'_';
			end
		end
		if (strlen(gemsString) ~= 0) then
			string = string..",gems="..strsub(gemsString, 0, strlen(gemsString)-1);
		end
	end

	
	return string
end

function SimcraftizeName(name) 
	if (name == nil) then return nil end
	return gsub(gsub(string.lower(name), "[^a-z ]", ""), " +", "_")
end